#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from base64 import b64encode
from unittest import mock

import pytest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.microsoft.winrm.operators.winrm import WinRMOperator


class TestWinRMOperator:
    def test_no_winrm_hook_no_ssh_conn_id(self):
        op = WinRMOperator(task_id="test_task_id", winrm_hook=None, ssh_conn_id=None)
        exception_msg = "Cannot operate without winrm_hook or ssh_conn_id."
        with pytest.raises(AirflowException, match=exception_msg):
            op.execute(None)

    @mock.patch("airflow.providers.microsoft.winrm.operators.winrm.WinRMHook")
    def test_no_command(self, mock_hook):
        op = WinRMOperator(task_id="test_task_id", winrm_hook=mock_hook, command=None)
        exception_msg = "No command specified so nothing to execute here."
        with pytest.raises(AirflowException, match=exception_msg):
            op.execute(None)

    @mock.patch("airflow.providers.microsoft.winrm.operators.winrm.WinRMHook")
    def test_default_returning_0_command(self, mock_hook):
        stdout = [b"O", b"K"]
        command = "not_empty"
        working_dir = "c:\\temp"
        mock_hook.run.return_value = (0, stdout, [])
        op = WinRMOperator(
            task_id="test_task_id", winrm_hook=mock_hook, command=command, working_directory=working_dir
        )
        execute_result = op.execute(None)
        assert execute_result == b64encode(b"".join(stdout)).decode("utf-8")
        mock_hook.run.assert_called_once_with(
            command=command,
            ps_path=None,
            output_encoding="utf-8",
            return_output=True,
            working_directory=working_dir,
        )

    @mock.patch("airflow.providers.microsoft.winrm.operators.winrm.WinRMHook")
    def test_default_returning_1_command(self, mock_hook):
        stderr = [b"K", b"O"]
        command = "not_empty"
        mock_hook.run.return_value = (1, [], stderr)
        op = WinRMOperator(task_id="test_task_id", winrm_hook=mock_hook, command=command)
        exception_msg = f"Error running cmd: {command}, return code: 1, error: KO"
        with pytest.raises(AirflowException, match=exception_msg):
            op.execute(None)

    @mock.patch("airflow.providers.microsoft.winrm.operators.winrm.WinRMHook")
    @pytest.mark.parametrize("expected_return_code", [1, [1, 2], range(1, 3)])
    @pytest.mark.parametrize("real_return_code", [0, 1, 2])
    def test_expected_return_code_command(self, mock_hook, expected_return_code, real_return_code):
        stdout = [b"O", b"K"]
        stderr = [b"K", b"O"]
        command = "not_empty"
        mock_hook.run.return_value = (real_return_code, stdout, stderr)
        op = WinRMOperator(
            task_id="test_task_id",
            winrm_hook=mock_hook,
            command=command,
            expected_return_code=expected_return_code,
        )

        should_task_succeed = False
        if isinstance(expected_return_code, int):
            should_task_succeed = real_return_code == expected_return_code
        elif isinstance(expected_return_code, list) or isinstance(expected_return_code, range):
            should_task_succeed = real_return_code in expected_return_code

        if should_task_succeed:
            execute_result = op.execute(None)
            assert execute_result == b64encode(b"".join(stdout)).decode("utf-8")
            mock_hook.run.assert_called_once_with(
                command=command,
                ps_path=None,
                output_encoding="utf-8",
                return_output=True,
                working_directory=None,
            )
        else:
            exception_msg = f"Error running cmd: {command}, return code: {real_return_code}, error: KO"
            with pytest.raises(AirflowException, match=exception_msg):
                op.execute(None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This is an example dag for using the WinRMOperator.
"""

from __future__ import annotations

# --------------------------------------------------------------------------------
# Load The Dependencies
# --------------------------------------------------------------------------------
import os
from datetime import datetime, timedelta

from airflow import DAG
from airflow.providers.microsoft.winrm.hooks.winrm import WinRMHook
from airflow.providers.microsoft.winrm.operators.winrm import WinRMOperator

# --------------------------------------------------------------------------------
# Caveat: This Dag will not run because of missing scripts.
# The purpose of this is to give you a sample of a real world example DAG!
# --------------------------------------------------------------------------------
from airflow.providers.standard.operators.empty import EmptyOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID")
DAG_ID = "POC_winrm_parallel"

with DAG(
    dag_id=DAG_ID,
    schedule="0 0 * * *",
    start_date=datetime(2021, 1, 1),
    dagrun_timeout=timedelta(minutes=60),
    tags=["example"],
    catchup=False,
) as dag:
    run_this_last = EmptyOperator(task_id="run_this_last")

    # [START create_hook]
    winRMHook = WinRMHook(ssh_conn_id="ssh_POC1")
    # [END create_hook]

    # [START run_operator]
    t1 = WinRMOperator(task_id="wintask1", command="ls -altr", winrm_hook=winRMHook)

    t2 = WinRMOperator(task_id="wintask2", command="sleep 60", winrm_hook=winRMHook)

    t3 = WinRMOperator(task_id="wintask3", command="echo 'luke test' ", winrm_hook=winRMHook)
    # [END run_operator]

    [t1, t2, t3] >> run_this_last

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.microsoft.winrm.hooks.winrm import WinRMHook

try:
    from airflow.sdk import Connection  # type: ignore
except ImportError:
    from airflow.models import Connection  # type: ignore


class TestWinRMHook:
    def test_get_conn_missing_remote_host(self):
        with pytest.raises(AirflowException):
            WinRMHook().get_conn()

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol")
    def test_get_conn_error(self, mock_protocol):
        mock_protocol.side_effect = Exception("Error")

        with pytest.raises(AirflowException):
            WinRMHook(remote_host="host", password="pwd").get_conn()

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol", autospec=True)
    @patch(
        "airflow.providers.microsoft.winrm.hooks.winrm.WinRMHook.get_connection",
        return_value=Connection(
            conn_id="",
            conn_type="",
            login="username",
            password="password",
            host="remote_host",
            extra="""{
                   "endpoint": "endpoint",
                   "remote_port": 123,
                   "transport": "plaintext",
                   "service": "service",
                   "keytab": "keytab",
                   "ca_trust_path": "ca_trust_path",
                   "cert_pem": "cert_pem",
                   "cert_key_pem": "cert_key_pem",
                   "server_cert_validation": "validate",
                   "kerberos_delegation": "true",
                   "read_timeout_sec": 124,
                   "operation_timeout_sec": 123,
                   "kerberos_hostname_override": "kerberos_hostname_override",
                   "message_encryption": "auto",
                   "credssp_disable_tlsv1_2": "true",
                   "send_cbt": "false"
               }""",
        ),
    )
    def test_get_conn_from_connection(self, mock_get_connection, mock_protocol):
        connection = mock_get_connection.return_value
        winrm_hook = WinRMHook(ssh_conn_id="conn_id")

        winrm_hook.get_conn()

        mock_get_connection.assert_called_once_with(winrm_hook.ssh_conn_id)
        mock_protocol.assert_called_once_with(
            endpoint=str(connection.extra_dejson["endpoint"]),
            transport=str(connection.extra_dejson["transport"]),
            username=connection.login,
            password=connection.password,
            service=str(connection.extra_dejson["service"]),
            keytab=str(connection.extra_dejson["keytab"]),
            ca_trust_path=str(connection.extra_dejson["ca_trust_path"]),
            cert_pem=str(connection.extra_dejson["cert_pem"]),
            cert_key_pem=str(connection.extra_dejson["cert_key_pem"]),
            server_cert_validation=str(connection.extra_dejson["server_cert_validation"]),
            kerberos_delegation=str(connection.extra_dejson["kerberos_delegation"]).lower() == "true",
            read_timeout_sec=int(connection.extra_dejson["read_timeout_sec"]),
            operation_timeout_sec=int(connection.extra_dejson["operation_timeout_sec"]),
            kerberos_hostname_override=str(connection.extra_dejson["kerberos_hostname_override"]),
            message_encryption=str(connection.extra_dejson["message_encryption"]),
            credssp_disable_tlsv1_2=str(connection.extra_dejson["credssp_disable_tlsv1_2"]).lower() == "true",
            send_cbt=str(connection.extra_dejson["send_cbt"]).lower() == "true",
        )

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.getuser", return_value="user")
    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol")
    def test_get_conn_no_username(self, mock_protocol, mock_getuser):
        winrm_hook = WinRMHook(remote_host="host", password="password")

        winrm_hook.get_conn()

        assert mock_getuser.return_value == winrm_hook.username

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol")
    def test_get_conn_no_endpoint(self, mock_protocol):
        winrm_hook = WinRMHook(remote_host="host", password="password")

        winrm_hook.get_conn()

        assert f"http://{winrm_hook.remote_host}:{winrm_hook.remote_port}/wsman" == winrm_hook.endpoint

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol", autospec=True)
    @patch(
        "airflow.providers.microsoft.winrm.hooks.winrm.WinRMHook.get_connection",
        return_value=Connection(
            conn_id="",
            conn_type="",
            login="username",
            password="password",
            host="remote_host",
            extra="""{
                      "endpoint": "endpoint",
                      "remote_port": 123,
                      "transport": "plaintext",
                      "service": "service",
                      "keytab": "keytab",
                      "ca_trust_path": "ca_trust_path",
                      "cert_pem": "cert_pem",
                      "cert_key_pem": "cert_key_pem",
                      "server_cert_validation": "validate",
                      "kerberos_delegation": "true",
                      "read_timeout_sec": 124,
                      "operation_timeout_sec": 123,
                      "kerberos_hostname_override": "kerberos_hostname_override",
                      "message_encryption": "auto",
                      "credssp_disable_tlsv1_2": "true",
                      "send_cbt": "false"
                  }""",
        ),
    )
    def test_run_with_stdout(self, mock_get_connection, mock_protocol):
        winrm_hook = WinRMHook(ssh_conn_id="conn_id")

        mock_protocol.return_value.run_command = MagicMock(return_value="command_id")
        mock_protocol.return_value.get_command_output_raw = MagicMock(
            return_value=(b"stdout", b"stderr", 0, True)
        )

        return_code, stdout_buffer, stderr_buffer = winrm_hook.run("dir")

        assert return_code == 0
        assert stdout_buffer == [b"stdout"]
        assert stderr_buffer == [b"stderr"]

    @patch("airflow.providers.microsoft.winrm.hooks.winrm.Protocol", autospec=True)
    @patch(
        "airflow.providers.microsoft.winrm.hooks.winrm.WinRMHook.get_connection",
        return_value=Connection(
            conn_id="",
            conn_type="",
            login="username",
            password="password",
            host="remote_host",
            extra="""{
                      "endpoint": "endpoint",
                      "remote_port": 123,
                      "transport": "plaintext",
                      "service": "service",
                      "keytab": "keytab",
                      "ca_trust_path": "ca_trust_path",
                      "cert_pem": "cert_pem",
                      "cert_key_pem": "cert_key_pem",
                      "server_cert_validation": "validate",
                      "kerberos_delegation": "true",
                      "read_timeout_sec": 124,
                      "operation_timeout_sec": 123,
                      "kerberos_hostname_override": "kerberos_hostname_override",
                      "message_encryption": "auto",
                      "credssp_disable_tlsv1_2": "true",
                      "send_cbt": "false"
                  }""",
        ),
    )
    def test_run_without_stdout_and_working_dir(self, mock_get_connection, mock_protocol):
        winrm_hook = WinRMHook(ssh_conn_id="conn_id")
        working_dir = "c:\\test"
        mock_protocol.return_value.run_command = MagicMock(return_value="command_id")
        mock_protocol.return_value.get_command_output_raw = MagicMock(
            return_value=(b"stdout", b"stderr", 0, True)
        )
        mock_protocol.return_value.open_shell = MagicMock()

        return_code, stdout_buffer, stderr_buffer = winrm_hook.run(
            "dir", return_output=False, working_directory=working_dir
        )

        mock_protocol.return_value.open_shell.assert_called_once_with(working_directory=working_dir)
        assert return_code == 0
        assert not stdout_buffer
        assert stderr_buffer == [b"stderr"]
